import org.graphstream.algorithm.generator.Generator;
import org.graphstream.algorithm.generator.RandomGenerator;
import org.graphstream.graph.Edge;
import org.graphstream.graph.Graph;
import org.graphstream.graph.Node;
import org.graphstream.graph.Path;
import org.graphstream.graph.implementations.SingleGraph;

import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Objects;

public class Dijkstra {

	private final static int TEST_GRAPH_COUNT = 25;
	private final static double AVERAGE_DEGREE = 3;
	private final static int NODE_COUNT = 500;
	private static final DecimalFormat df = new DecimalFormat("0.00");

	/**
	 * Test method, will create a defined number of graphs and compare the results of the custom algorithm to the one impllemented in graphstream
	 * @param args arguments
	 */
	public static void main(String[] args) {
		Graph[] graphs = new Graph[TEST_GRAPH_COUNT];
		for(int i = 0; i < TEST_GRAPH_COUNT; ++i) {
			graphs[i] = Dijkstra.getRandomGraph("random" + String.format("%03d", i), AVERAGE_DEGREE, NODE_COUNT);
		}

		ArrayList<Path> resultsIntegrated = new ArrayList<>();
		ArrayList<Path> resultsCustom     = new ArrayList<>();

		ArrayList<Duration> timesIntegrated = new ArrayList<>();
		ArrayList<Duration> timesCustom     = new ArrayList<>();

		ArrayList<Double> distIntegrated = new ArrayList<>();
		ArrayList<Double> distCustom      = new ArrayList<>();

		// For simplicity, source will always be 0 and destination the last node
		for(Graph g : graphs) {
			// Initialisation of source and dest
			Node source = g.getNode(0);
			Node dest = g.getNode(NODE_COUNT-1);

			// Initialisation of the from library Dijkstra algorithm
			org.graphstream.algorithm.Dijkstra d = new org.graphstream.algorithm.Dijkstra(org.graphstream.algorithm.Dijkstra.Element.EDGE, "result", "length");
			d.init(g);
			d.setSource(source);

			// Start calculating and measure of the time it takes
			Instant start = Instant.now();
			d.compute();
			Instant end = Instant.now();

			// Start of the custom algorithm and time meaurement
			Instant startCustom = Instant.now();
			DijkstraResult resultCustom = dijkstra(g, source);
			Instant endCustom = Instant.now();

			// Sauvegarde des résultats
			//   Paths
			resultsIntegrated.add(d.getPath(dest));
			resultsCustom.add(resultCustom.getPath(dest));
			//   Distances
			distIntegrated.add(d.getPathLength(dest));
			distCustom.add(resultCustom.getDistance(dest));

			// Sauvegarde des temps
			timesIntegrated.add(Duration.between(start, end));
			timesCustom.add(Duration.between(startCustom, endCustom));
		}

		System.out.printf("%25s | %25s\n", "Times integrated (ms)", "Times custom (ms)");
		for(int i = 0; i < TEST_GRAPH_COUNT; ++i) {
			System.out.printf("%25d | %25d\n", timesIntegrated.get(i).toMillis(), timesCustom.get(i).toMillis());
		}

		System.out.println("\n_________________________________________________________________\n");

		System.out.printf("%60s | %-60s\n", "Results integrated", "Times custom");
		for(int i = 0; i < TEST_GRAPH_COUNT; ++i) {
			System.out.printf("%60s | %-60s\n", resultsIntegrated.get(i).toString(), resultsCustom.get(i).toString());
		}

		System.out.println("\n_________________________________________________________________\n");

		System.out.printf("%20s | %20s\n", "Dist integrated", "Dist custom");
		for(int i = 0; i < TEST_GRAPH_COUNT; ++i) {
			System.out.printf("%20f | %20f\n", distIntegrated.get(i), distCustom.get(i));
		}

		System.out.println("\n_________________________________________________________________\n");

		ArrayList<Integer> differentPaths = new ArrayList<>();
		ArrayList<Integer> differentDists = new ArrayList<>();
		for(int i = 0; i < TEST_GRAPH_COUNT; ++i) {
			if(!resultsCustom.get(i).equals(resultsIntegrated.get(i))) differentPaths.add(i);
			if(!distCustom.get(i).equals(distIntegrated.get(i))) differentDists.add(i);
		}

		if(differentPaths.isEmpty())
			System.out.println("No differences between integrated and custom version on resulted paths");
		else
			System.out.println("There is a difference between integrated and custom resulted paths on these graphs : " + differentPaths);

		if(differentDists.isEmpty())
			System.out.println("No differences between integrated and custom version on resulted distances");
		else
			System.out.println("There is a difference between integrated and custom distances on these graphs : " + differentDists);

		System.out.println("The time difference between the integrated version and the custom version is averaging " + df.format(((double)timesCustom.stream().map(Duration::getNano).reduce(Integer::sum).orElseThrow() / timesIntegrated.stream().map(Duration::getNano).reduce(Integer::sum).orElseThrow()) * 100) + "%");

		try {
			FileWriter fw = new FileWriter("./gnuplot_data/int"+NODE_COUNT+".dat");
			FileWriter fwcus = new FileWriter("./gnuplot_data/cus"+NODE_COUNT+".dat");
			StringBuilder sb = new StringBuilder();
			StringBuilder sbcus = new StringBuilder();
			for(int i = 0; i < timesIntegrated.size(); ++i) {
				sb.append(i).append(" ").append(timesIntegrated.get(i).toMillis()).append("\n");
				sbcus.append(i).append(" ").append(timesCustom.get(i).toMillis()).append("\n");
			}
			fw.write(sb.toString());
			fw.close();
			fwcus.write(sbcus.toString());
			fwcus.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param g The graph on which you want to run dijkstra algorithm on
	 * @param source The source node from which you want to calculate distances
	 * @return An object of type DijkstraResult on which you can require path
	 */
	private static DijkstraResult dijkstra(Graph g, Node source) {
		assert source != null;
		HashMap<Node, Node> previous = new HashMap<>();
		HashMap<Node, Double> distances = new HashMap<>();
		ArrayList<Node> nodes = new ArrayList<>();
		g.nodes().forEach(n -> {
			nodes.add(n);
			distances.put(n, Double.POSITIVE_INFINITY);
			previous.put(n, null);
		});
		distances.put(source, 0.0);
		while(!nodes.isEmpty()) {
			Node n = findMin(nodes, distances);
			if(n == null) break;
			nodes.remove(n);
			n.neighborNodes().forEach(nei -> updateDistances(distances, previous, n, nei));
		}
		return new DijkstraResult(previous, distances, source);
	}

	/**
	 * @param nodes ArrayList of nodes where you want to find the minimum distance
	 * @param distances the HashMap of distances
	 * @return the Node with the minimum distance
	 */
	private static Node findMin(ArrayList<Node> nodes, HashMap<Node, Double> distances) {
		Double min = Double.POSITIVE_INFINITY;
		Node minNode = null;
		for(Node n : nodes) {
			Double dist = distances.get(n);
			if(dist < min) {
				min = dist;
				minNode = n;
			}
			if(min == 0.0) break;
		}
		return minNode;
	}

	/**
	 * @param distances the HashMap of distances
	 * @param previous the HashMap of previous nodes
	 * @param n1 Node number 1
	 * @param n2 Node number 2
	 */
	private static void updateDistances(HashMap<Node, Double> distances, HashMap<Node, Node> previous, Node n1, Node n2) {
		Double distN1 = distances.get(n1);
		Double distN2 = distances.get(n2);
		Edge n1n2 = n1.getEdgeBetween(n2);
		assert n1n2 != null;
		Double lengthN1N2 = Double.parseDouble(n1n2.getAttribute("length").toString());
		if(distN2 > distN1 + lengthN1N2) {
			distances.put(n2, distances.get(n1) + lengthN1N2);
			previous.put(n2, n1);
		}
	}

	/**
	 * Creates a random graph with a random value 'length' for each edge
	 * @param id the id of the random graph
	 * @param averageDegree the average degree of this graph
	 * @param nodeCount the number of created nodes
	 * @return a random SingleGraph
	 */
	private static Graph getRandomGraph(String id, double averageDegree, int nodeCount) {
		Graph g = new SingleGraph(id);
		Generator gen = new RandomGenerator(averageDegree, false, false, "", "length");
		gen.addSink(g);
		gen.begin();
		for(int i = 0; i < nodeCount; ++i)
			gen.nextEvents();
		gen.end();
		gen.removeSink(g);
		return g;
	}

	/**
	 * Class used with the algorithm to store the result
	 */
	private static class DijkstraResult {
		HashMap<Node, Node> previous;
		HashMap<Node, Double> distances;
		Node source;
		DijkstraResult(HashMap<Node, Node> previous, HashMap<Node, Double> distances, Node source) {
			this.previous = previous;
			this.distances = distances;
			this.source = source;
		}

		/**
		 * Returns the path between the source node used in the algorithm and the destination
		 * @param destination the Noode you want to get the path to
		 * @return the shortest Path between the source and destination node
		 */
		Path getPath(Node destination) {
			ArrayList<Edge> pathList = new ArrayList<>();
			Node pathNode = destination;
			while(pathNode != source) {
				if(pathNode == null) break;
				Edge e = pathNode.getEdgeBetween(previous.get(pathNode));
				pathList.add(e);
				pathNode = previous.get(pathNode);
			}
			if(pathList.stream().allMatch(Objects::isNull)) return new Path();
			Collections.reverse(pathList);
			Path path = new Path();
			path.setRoot(source);
			for(Edge e : pathList) {
				path.add(e);
			}
			return path;
		}

		/**
		 * Gives the distance between the source node and the given node
		 * @param destination The node you want to get the distance from the source
		 * @return the distance between the source and destination
		 */
		Double getDistance(Node destination) {
			return distances.get(destination);
		}
	}
}
